      subroutine cmpt_ctf(fft_klh, hkl_max, fft_ann,
     $     ctf_val, abs_ctf, disef, fno, weight, nptles,
     $     pixels, mag, theta, phi, omega, x, y,
     $     defocus_max, defocus_min, defocus_ang,
     $     score, orient_flag, origin_flag)

C     This routine searches 'defocus space' to determine the major
C     defocus, minor defocus,and astigmatism orientation that results in
C     the best overall agreement between the particle images in a
C     micrograph and the model.
C
C     Although the image/model comparison can be done using any choice
C     of funcmode, the phase method (funcmode=3) works best since it
C     just uses the phase information to determine the optimal location
C     of the CTF nodes.

      implicit none

      include '../include/sizes.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'

      integer, intent(in) :: hkl_max, fft_ann, fno, nptles
      real, intent(in)    :: weight
      complex, intent(in) :: fft_klh(-fft_ann:fft_ann, -fft_ann:fft_ann, 
     $     0:fft_ann)
      real, intent(out) :: ctf_val(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(out) :: abs_ctf(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in)  :: disef(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)
      real, intent(in)  :: pixels(npix_row, npix_col)

      real, intent(in)    :: theta, phi, omega, x, y, mag
      real, intent(inout) :: defocus_max, defocus_min, defocus_ang
      real, intent(out)   :: score
     $                (-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle)
      character(len=1), intent(out) :: orient_flag, origin_flag

      integer defmaj_step, defmaj_nangle
      integer defmin_step, defmin_nangle
      integer angle_step,  angle_nangle

      integer i, j, k, indx1, indx2
      integer :: ishift, maxshift = 10
      integer min_ctfs(3)
      integer alloc_stat, tmp_nangle

      real :: imgmagfactor
      real :: r4_max = huge(r4_max)
      real du, dv
      real center(2),min_value,
     +     num_annul, value, min_annul, ctfs(3)
      double precision part_dist
      real    rot_mat(3,3), 
     +     dist(-Nangle:Nangle,-Nangle:Nangle,-Nangle:Nangle)

      real, allocatable :: rffta_omega(:)
      complex, allocatable :: fc_2d_ref(:,:),fc_2d_intp(:,:),fc_2d(:,:)
      
      character(len=12) :: subname = 'cmpt_ctfr', progname ='PCTFR'

C     Avoid calculations on ptles with terrible origins
      if(  x < npix_row*0.1 .or. x > npix_row*0.9 .or.
     $     y < npix_col*0.1 .or. y > npix_col*0.9) then
         score = 9999.999              !F90 array syntax
         orient_flag = 'X'
         origin_flag = 'X'
         return
      endif

      orient_flag = 'S'
      origin_flag = 'S'
      allocate(rffta_omega(nrffta_omega),
     +     fc_2d_ref(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col), 
     +     fc_2d_intp(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     fc_2d(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col),
     +     stat = alloc_stat)
      if(alloc_stat > 0) then
         call error_stop (progname, subname, 
     +        ' Memory allocation failed')
      endif

C     Set loop bounds for stigmatic and astigmatic cases
      defmaj_nangle = nangle
      defmin_nangle = nangle
      angle_nangle  = nangle

C     to initialize the value of rffta_omega(nrfft_col+15) that is
C     rffta_omega(nrffta_omega) Initialize the FFT parameters used to
C     transform rows of RFFTA each row has length NRFFTA_COL
      call vrffti( nrffta_col, rffta_omega )

      min_value = R4_MAX
      score = R4_MAX              ! f90 array syntax
      
      num_annul = real(nrffta_row)*angstrom_per_pix/res_max 
      min_annul = real(nrffta_row)*angstrom_per_pix/res_min
      if (num_annul >= nrffta_row/2.) 
     +     call error_stop(progname,subname, 
     +     'Resolution is too high, smaller than 2*angstrom_per_pix')

      center(1)    = x
      center(2)    = y
      imgmagfactor = mag  
      
            
C     adjust the imgmagfactor according to its own pixel_size
      imgmagfactor = imgmagfactor * PIXSIZ_OUT / angstrom_per_pix

C     Compute Discrete Fourier Transform of pixel values:
      call fft_2dfft(pixels, npix_row, npix_col, center, nrffta_omega,
     $     rffta_omega, nrffta_row, nrffta_col, nfc_2d_row, nfc_2d_col,
     $     fc_2d )

C     Need to apply tempfac to images
      call maptempfac(fc_2d, 0, nfc_2d_row, -nfc_2d_col,
     $     nfc_2d_col, 1, 1, tempfac, pixsiz_out, units)
      
      du = center(1) - real(int(center(1)))
      dv = center(2) - real(int(center(2)))
      call shift_fft(fc_2d, nfc_2d_row, nfc_2d_col, du, dv)

C     First calculate rotation matrix and extract slice
      call crowther_to_matrix(theta, -phi, -omega, rot_mat)

      call interpl_3d(fft_klh, hkl_max, min_annul, num_annul, rot_mat
     $        , fc_2d_intp, imgmagfactor, fft_ann)

C     loop to refine CTF, use min_ctfs and ctfs as temp value of CTF
C     the resolution for defocus is Dangle, and the resolution for angle
C     is Dangle*10

      
      if (per_ptle_ctf == 0) then
         maxshift = 1
C this assignment should be taken care of in the calling routine         
         defocus_max = DELF_MAJS(fno)
         defocus_min = DELF_MINS(fno)
         defocus_ang = ANG_MAJS(fno)
      endif
      do ishift = 1, maxshift
               
        min_ctfs(1) = 0           ! defocus major
        min_ctfs(2) = 0           ! defocus minor
        min_ctfs(3) = 0           ! the angle
 
        do defmaj_step = -defmaj_nangle, defmaj_nangle  
          ctfs(1)= dangle*defmaj_step + defocus_max
          if (ctfs(1) <= 0) cycle ! Skip to next iteration
          do defmin_step = -defmin_nangle, defmin_nangle  
            ctfs(2)= dangle*defmin_step + defocus_min
            if (ctfs(2) <= 0) cycle ! Skip to next iteration

C     If ctfs(1) equals ctfs(2), then don't need to rotate about
C     angle. To avoid having to reproduce a big chunk of code, set the
C     loop indices based on the result of this test.
            if (ctfs(2) == ctfs(1)) then
               tmp_nangle = 0
            else
               tmp_nangle = angle_nangle
            endif

            do angle_step = -tmp_nangle, tmp_nangle
               ctfs(3) = dangle*angle_step*10 + defocus_ang
               if (score(defmaj_step,defmin_step,angle_step) < 
     +              R4_MAX-1) cycle
               call ctf_para(ctf_val, abs_ctf, 
     +              0, nfc_2d_row, -nfc_2d_col, nfc_2d_col, .true.,
     +              CTFMODE, Filter, TEMPFAC, CTF_FF1, CTF_FF2,
     +              PIXSIZS(fno), units, VOLTS(fno),ctfs(1), 
     +              ctfs(2), ctfs(3), AMPFACS(fno),
     +              CS_COEFFS(fno), Weight, res_max, res_max)

               fc_2d_ref = fc_2d_intp * ctf_val

               call imgcomport(fc_2d_ref, hkl_max, fc_2d, disef,
     +               abs_ctf, min_annul, num_annul, 
     +               score(defmaj_step,defmin_step,angle_step),
     +               funcmode)

            enddo
          enddo
        enddo
      
C     Determine location of minimum element in array score. Note use
C     of Fortran 90 array syntax and intrinsic function minloc (which
C     behaves as if lower bounds of array are all equal to 1)

        min_ctfs = minloc(score)
        min_ctfs = min_ctfs - nangle - 1

        defocus_max = defocus_max + min_ctfs(1) * Dangle
        defocus_min = defocus_min + min_ctfs(2) * Dangle
        defocus_ang = defocus_ang + min_ctfs(3) * Dangle * 10

        if ( (per_ptle_ctf == 1) .and. 
     +     (abs(real(min_ctfs(1))) > Nangle*0.5 .or. 
     +     abs(real(min_ctfs(2))) > Nangle*0.5 .or.
     +     abs(real(min_ctfs(3))) > Nangle*0.5) ) then
         
C     slide score window to move element with min value in center
          call recenter_array_3d(score, Nangle, min_ctfs, R4_MAX)
        endif

      enddo

      
      deallocate(rffta_omega, fc_2d, fc_2d_intp, fc_2d_ref)

      return
      end subroutine cmpt_ctf
